# coding: utf-8
"""Aliases management"""

from sys import argv
from yaml import load, error
from stakkr.configreader import get_config_and_project_dir


def get_config_from_argv(argv: list):
    """Search for a config file option in command line"""
    for index, arg in enumerate(argv):
        # manage "=" sign
        if arg.find('=') != -1 and arg.split('=')[0] == '--config':
            return arg.split('=')[1]

        # else search for -c  or --config and get the next value
        if arg in ['-c', '--config']:
            try:
                return argv[index + 1]
            except IndexError:
                raise ValueError('You must specify the config file name')

    return None


def get_aliases():
    """Get aliases from config file"""
    config_file, _ = get_config_and_project_dir(get_config_from_argv(argv[1:]))
    config = {}
    try:
        with open(config_file, 'r') as stream:
            config = load(stream)
    except (error.YAMLError, FileNotFoundError):
        pass

    return config['aliases'] if 'aliases' in config else {}



#!/usr/bin/env python
# coding: utf-8
"""
CLI Main Entry Point.

Wraps docker-compose and build it from what has been taken from config.
"""

import glob
import os
import subprocess
import sys
import click
from stakkr import file_utils
from stakkr.configreader import Config


@click.command(help="Wrapper for docker-compose",
               context_settings=dict(ignore_unknown_options=True))
@click.option('--config-file', '-c', help="Set stakkr config file location (default stakkr.yml)")
@click.argument('command', nargs=-1, type=click.UNPROCESSED)
def cli(config_file: str = None, command: tuple = ()):
    """Command line entry point."""
    config, config_file = _get_config(config_file)

    # Set main config and services as env variables
    _set_env_from_config(config)

    # Register proxy env parameters
    _set_env_for_proxy(config['proxy'])

    # set the base command
    base_cmd = _get_base_command(config)

    msg = click.style('[VERBOSE] ', fg='green')
    msg += 'Compose command: ' + ' '.join(base_cmd + list(command))
    click.echo(msg, err=True)
    subprocess.call(base_cmd + list(command))


def _add_local_services(project_dir: str, available_services: list):
    """Get services in the virtualenv services/ directory, so specific to that stakkr."""
    services_dir = project_dir + '/services/*/docker-compose'
    for service_dir in glob.glob(services_dir):
        conf_files = _get_services_from_dir(service_dir)
        for conf_file in conf_files:
            available_services[conf_file[:-4]] = service_dir + '/' + conf_file

    return available_services


def get_available_services(project_dir: str):
    """Get standard services bundled with stakkr."""
    services_dir = file_utils.get_dir('static') + '/services/'
    conf_files = _get_services_from_dir(services_dir)

    services = dict()
    for conf_file in conf_files:
        services[conf_file[:-4]] = services_dir + conf_file

    services = _add_local_services(project_dir, services)

    return services


def _get_base_command(config: dict):
    """Build the docker-compose file to be run as a command."""
    main_file = 'docker-compose.yml'
    # Set the network subnet ?
    if config['subnet'] != '':
        main_file = 'docker-compose.subnet.yml'
    cmd = ['docker-compose', '-f', file_utils.get_file('static', main_file)]

    # What to load
    activated_services = _get_enabled_services_files(
        config['project_dir'],
        [svc for svc, opts in config['services'].items() if opts['enabled'] is True])
    # Create the command
    services = []
    for service in activated_services:
        services.append('-f')
        services.append(service)

    return cmd + services + ['-p', config['project_name']]


def _get_config(config_file: str):
    """Read main stakkr.yml file."""
    config_reader = Config(config_file)
    config = config_reader.read()

    if config is False:
        config.display_errors()
        sys.exit(1)

    return config, config_reader.config_file


def _get_enabled_services_files(project_dir: str, configured_services: list):
    """Compile all available services : standard and local install."""
    available_services = get_available_services(project_dir)

    services_files = []
    for service in configured_services:
        if service not in available_services:
            msg = 'Error: service "{}" has no configuration file. '.format(service)
            msg += 'Check your config'
            click.secho(msg, fg='red')
            sys.exit(1)
        services_files.append(available_services[service])

    return services_files


def _get_gid(gid: int):
    if gid is not None:
        return str(gid)

    return '1000' if os.name == 'nt' else str(os.getgid())


def _get_services_from_dir(services_dir: str):
    if os.path.isdir(services_dir) is False:
        return []

    return [service for service in os.listdir(services_dir) if service.endswith('.yml')]


def _get_uid(uid: int):
    if uid is not None:
        return str(uid)

    return '1000' if os.name == 'nt' else str(os.getuid())


def _set_env_for_proxy(config: dict):
    """Define environment variables to be used in services yaml."""
    os.environ['PROXY_ENABLED'] = str(config['enabled'])
    os.environ['PROXY_DOMAIN'] = str(config['domain'])


def _set_env_for_services(services: dict):
    for service, params in services.items():
        if params['enabled'] is False:
            continue

        for param, value in params.items():
            env_var = 'DOCKER_{}_{}'.format(service, param).upper()
            os.environ[env_var] = str(value)


def _set_env_from_config(config: dict):
    """Define environment variables to be used in services yaml."""
    os.environ['COMPOSE_BASE_DIR'] = config['project_dir']
    os.environ['COMPOSE_PROJECT_NAME'] = config['project_name']
    for parameter, value in config.items():
        if parameter == 'services':
            _set_env_for_services(value)
            continue

        os.environ['DOCKER_{}'.format(parameter.upper())] = str(value)

    # Do that at the end, else the value in config will overwrite the possible
    # default value
    os.environ['DOCKER_UID'] = _get_uid(config['uid'])
    os.environ['DOCKER_GID'] = _get_gid(config['gid'])


if __name__ == '__main__':
    cli()



#!/usr/bin/env python
# coding: utf-8
"""
Docker Clean command.

Clean unused containers, images, volumes and networks.
That saves a lot of space ...
"""

import sys
import click
import humanfriendly
from stakkr.docker_actions import get_client as get_docker_client


@click.command(help="""Clean Docker containers, images, volumes and networks
that are not in use""", name="docker-clean")
@click.option('--force', '-f', help="Do it", is_flag=True)
@click.option('--containers/--no-containers', '-c/', help="Remove containers", is_flag=True, default=True)
@click.option('--images/--no-images', '-i/', help="Remove images", is_flag=True, default=True)
@click.option('--volumes/--no-volumes', '-V/', help="Remove volumes", is_flag=True, default=True)
@click.option('--networks/--no-networks', '-n/', help="Remove networks", is_flag=True, default=True)
def clean(
        force: bool = False,
        containers: bool = True,
        images: bool = True,
        volumes: bool = True,
        networks: bool = True):
    """See command help."""

    if containers is True:
        click.secho('Cleaning Docker stopped containers:', fg='green')
        remove_containers(force)
    if images is True:
        click.secho('Cleaning Docker unused images:', fg='green')
        remove_images(force)
    if volumes is True:
        click.secho('Cleaning Docker unused volumes:', fg='green')
        remove_volumes(force)
    if networks is True:
        click.secho('Cleaning Docker unused networks:', fg='green')
        remove_networks(force)


def remove_containers(force: bool):
    """Remove exited containers."""
    stopped_containers = get_docker_client().containers.list(filters={'status': 'exited'})
    if not stopped_containers:
        print('  No exited container to remove')
        return

    if force is False:
        click.secho("  --force is not set so I won't do anything", fg='red')
        return

    res = get_docker_client().containers.prune()
    cts = len(res['ContainersDeleted'])
    space = humanfriendly.format_size(res['SpaceReclaimed'])
    click.echo('  Removed {} exited container(s), saved {}'.format(cts, space))


def remove_images(force: bool):
    """Remove unused images."""
    if force is False:
        click.secho("  --force is not set so I won't do anything", fg='red')
        return

    res = get_docker_client().images.prune(filters={'dangling': False})
    if res['ImagesDeleted'] is None:
        print('  No image to remove')
        return

    images = len(res['ImagesDeleted'])
    space = humanfriendly.format_size(res['SpaceReclaimed'])
    click.echo('  Removed {} images(s), saved {}'.format(images, space))


def remove_networks(force: bool):
    """Remove unused networks."""
    if force is False:
        click.secho("  --force is not set so I won't do anything", fg='red')
        return

    res = get_docker_client().networks.prune()
    if res['NetworksDeleted'] is None:
        print('  No network to remove')
        return

    networks = len(res['NetworksDeleted'])
    click.echo('  Removed {} network(s)'.format(networks))


def remove_volumes(force: bool):
    """Remove unused volumes."""
    if force is False:
        click.secho("  --force is not set so I won't do anything", fg='red')
        return

    res = get_docker_client().volumes.prune()
    if (res['VolumesDeleted'] is None) or (not res['VolumesDeleted']):
        print('  No volume to remove')
        return

    volumes = len(res['VolumesDeleted'])
    space = humanfriendly.format_size(res['SpaceReclaimed'])
    click.echo('  Removed {} volume(s), saved {}'.format(volumes, space))


def main():
    """Generate the CLI."""
    try:
        clean()
    except Exception as error:
        msg = click.style(r""" ______ _____  _____   ____  _____
|  ____|  __ \|  __ \ / __ \|  __ \
| |__  | |__) | |__) | |  | | |__) |
|  __| |  _  /|  _  /| |  | |  _  /
| |____| | \ \| | \ \| |__| | | \ \
|______|_|  \_\_|  \_\\____/|_|  \_\

""", fg='yellow')
        msg += click.style('{}'.format(error), fg='red')

        click.echo(msg)
        click.echo()
        # raise error
        sys.exit(1)


if __name__ == '__main__':
    main()



# coding: utf-8
"""Manage public proxy to expose containers."""

import click
from docker.errors import DockerException
from stakkr import docker_actions as docker
from stakkr.file_utils import get_dir


class Proxy:
    """Main class that does actions asked by the cli."""

    def __init__(self, http_port: int = 80, https_port: int = 443, ct_name: str = 'proxy_stakkr'):
        """Set the right values to start the proxy."""
        self.ports = {'http': http_port, 'https': https_port}
        self.ct_name = ct_name
        self.docker_client = docker.get_client()

    def start(self, stakkr_network: str = None):
        """Start stakkr proxy if stopped."""
        if docker.container_running(self.ct_name) is False:
            print(click.style('[STARTING]', fg='green') + ' traefik')
            self._start_container()

        # Connect it to network if asked
        if stakkr_network is not None:
            docker.add_container_to_network(self.ct_name, stakkr_network)

    def stop(self):
        """Stop stakkr proxy."""
        if docker.container_running(self.ct_name) is False:
            return

        print(click.style('[STOPPING]', fg='green') + ' traefik')
        proxy_ct = self.docker_client.containers.get(self.ct_name)
        proxy_ct.stop()

    def _start_container(self):
        """Start proxy."""
        proxy_conf_dir = get_dir('static/proxy')
        try:
            self.docker_client.images.pull('traefik:latest')
            self.docker_client.containers.run(
                'traefik:latest', remove=True, detach=True,
                hostname=self.ct_name, name=self.ct_name,
                volumes=[
                    '/var/run/docker.sock:/var/run/docker.sock',
                    '{}/traefik.toml:/etc/traefik/traefik.toml'.format(proxy_conf_dir),
                    '{}/ssl:/etc/traefik/ssl'.format(proxy_conf_dir)],
                ports={80: self.ports['http'], 8080: 8080, 443: self.ports['https']})
        except DockerException as error:
            raise RuntimeError("Can't start proxy ...({})".format(error))



#!/usr/bin/env python
# coding: utf-8
"""
CLI Entry Point.

From click, build stakkr.
Give all options to manage services to be launched, stopped, etc.
"""

import sys
import click
from click.core import Context
from stakkr.docker_actions import get_running_containers_names
from stakkr.aliases import get_aliases


@click.group(help="""Main CLI Tool that easily create / maintain
a stack of services, for example for web development.

Read the configuration file and setup the required services by
linking and managing everything for you.""")
@click.version_option('4.0.5')
@click.option('--config', '-c', help='Set the configuration filename (stakkr.yml by default)')
@click.option('--debug/--no-debug', '-d', default=False)
@click.option('--verbose', '-v', is_flag=True)
@click.pass_context
def stakkr(ctx: Context, config=None, debug=False, verbose=True):
    """Click group, set context and main object."""
    from stakkr.actions import StakkrActions

    ctx.obj['CONFIG'] = config
    ctx.obj['DEBUG'] = debug
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['STAKKR'] = StakkrActions(ctx.obj)


@stakkr.command(help="""Enter a container to perform direct actions such as
install packages, run commands, etc.""")
@click.argument('container', required=True)
@click.option('--user', '-u', help="User's name. Valid choices : www-data or root")
@click.option('--tty/--no-tty', '-t/ ', is_flag=True, default=True, help="Use a TTY")
@click.pass_context
def console(ctx: Context, container: str, user: str, tty: bool):
    """See command Help."""
    ctx.obj['STAKKR'].init_project()
    ctx.obj['CTS'] = get_running_containers_names(ctx.obj['STAKKR'].project_name)
    if ctx.obj['CTS']:
        ct_choice = click.Choice(ctx.obj['CTS'])
        ct_choice.convert(container, None, ctx)

    ctx.obj['STAKKR'].console(container, _get_cmd_user(user, container), tty)


@stakkr.command(help="""Execute a command into a container.

Examples:\n
- ``stakkr -v exec mysql mysqldump -p'$MYSQL_ROOT_PASSWORD' mydb > /tmp/backup.sql``\n
- ``stakkr exec php php -v`` : Execute the php binary in the php container with option -v\n
- ``stakkr exec apache service apache2 restart``\n
""", name='exec', context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False))
@click.pass_context
@click.option('--user', '-u', help="User's name. Be careful, each container have its own users.")
@click.option('--tty/--no-tty', '-t/ ', is_flag=True, default=True, help="Use a TTY")
@click.argument('container', required=True)
@click.argument('command', required=True, nargs=-1, type=click.UNPROCESSED)
def exec_cmd(ctx: Context, user: str, container: str, command: tuple, tty: bool):
    """See command Help."""
    ctx.obj['STAKKR'].init_project()
    ctx.obj['CTS'] = get_running_containers_names(ctx.obj['STAKKR'].project_name)
    if ctx.obj['CTS']:
        click.Choice(ctx.obj['CTS']).convert(container, None, ctx)

    ctx.obj['STAKKR'].exec_cmd(container, _get_cmd_user(user, container), command, tty)


@stakkr.command(help="Restart all (or a single as CONTAINER) container(s)")
@click.argument('container', required=False)
@click.option('--pull', '-p', help="Force a pull of the latest images versions", is_flag=True)
@click.option('--recreate', '-r', help="Recreate all containers", is_flag=True)
@click.option('--proxy/--no-proxy', '-P', help="Restart the proxy", default=True)
@click.pass_context
def restart(ctx: Context, container: str, pull: bool, recreate: bool, proxy: bool):
    """See command Help."""
    print(click.style('[RESTARTING]', fg='green') + ' your stakkr services')
    try:
        ctx.invoke(stop, container=container, proxy=proxy)
    except Exception:
        pass

    ctx.invoke(start, container=container, pull=pull, recreate=recreate, proxy=proxy)


@stakkr.command(help="""List available services available for stakkr.yml
(with info if the service is enabled)""")
@click.pass_context
def services(ctx):
    """See command Help."""
    ctx.obj['STAKKR'].init_project()

    from stakkr.stakkr_compose import get_available_services

    print('Available services usable in stakkr.yml ', end='')
    print('({} = disabled) : '.format(click.style('✘', fg='red')))

    svcs = ctx.obj['STAKKR'].get_config()['services']
    enabled_svcs = [svc for svc, opts in svcs.items() if opts['enabled'] is True]
    available_svcs = get_available_services(ctx.obj['STAKKR'].project_dir)
    for available_svc in sorted(list(available_svcs.keys())):
        sign = click.style('✘', fg='red')
        if available_svc in enabled_svcs:
            version = svcs[available_svc]['version']
            sign = click.style(str(version), fg='green')

        print('  - {} ({})'.format(available_svc, sign))


@stakkr.command(help="""Download a pack of services from github (see github) containing services.
PACKAGE is the git url or package name.
NAME is the directory name to clone the repo.
""")
@click.argument('package', required=True)
@click.argument('name', required=False)
@click.pass_context
def services_add(ctx: Context, package: str, name: str):
    """See command Help."""
    from stakkr.services import install

    project_dir = _get_project_dir(ctx.obj['CONFIG'])
    services_dir = '{}/services'.format(project_dir)
    name = package if name is None else name
    success, message = install(services_dir, package, name)
    if success is False:
        click.echo(click.style(message, fg='red'))
        sys.exit(1)

    stdout_msg = 'Package "' + click.style(package, fg='green') + '" installed successfully'
    if message is not None:
        stdout_msg = click.style(message, fg='yellow')

    click.echo(stdout_msg)
    click.echo('Try ' + click.style('stakkr services', fg='green') + ' to see new available services')


@stakkr.command(help="Update all services packs in services/")
@click.pass_context
def services_update(ctx):
    """See command Help."""
    from stakkr.services import update_all

    project_dir = _get_project_dir(ctx.obj['CONFIG'])
    services_dir = '{}/services'.format(project_dir)
    update_all(services_dir)
    print(click.style('Packages updated', fg='green'))


@stakkr.command(help="Start all (or a single as CONTAINER) container(s) defined in compose.ini")
@click.argument('container', required=False)
@click.option('--pull', '-p', help="Force a pull of the latest images versions", is_flag=True)
@click.option('--recreate', '-r', help="Recreate all containers", is_flag=True)
@click.option('--proxy/--no-proxy', '-P', help="Start proxy", default=True)
@click.pass_context
def start(ctx: Context, container: str, pull: bool, recreate: bool, proxy: bool):
    """See command Help."""
    print(click.style('[STARTING]', fg='green') + ' your stakkr services')

    ctx.obj['STAKKR'].start(container, pull, recreate, proxy)
    _show_status(ctx)


@stakkr.command(help="Display a list of running containers")
@click.pass_context
def status(ctx):
    """See command Help."""
    ctx.obj['STAKKR'].status()


@stakkr.command(help="Stop all (or a single as CONTAINER) container(s)")
@click.argument('container', required=False)
@click.option('--proxy/--no-proxy', '-P', help="Stop the proxy", default=True)
@click.pass_context
def stop(ctx: Context, container: str, proxy: bool):
    """See command Help."""
    print(click.style('[STOPPING]', fg='yellow') + ' your stakkr services')
    ctx.obj['STAKKR'].stop(container, proxy)


def _get_cmd_user(user: str, container: str):
    users = {'apache': 'www-data', 'nginx': 'www-data', 'php': 'www-data'}

    cmd_user = 'root' if user is None else user
    if container in users and user is None:
        cmd_user = users[container]

    return cmd_user


def _show_status(ctx):
    services_ports = ctx.obj['STAKKR'].get_services_urls()
    if services_ports == '':
        print('\nServices Status:')
        ctx.invoke(status)
        return

    print('\nServices URLs :')
    print(services_ports)


def _get_project_dir(config: str):
    from os.path import abspath, dirname

    if config is not None:
        config = abspath(config)
        return dirname(config)

    from stakkr.file_utils import find_project_dir
    return find_project_dir()


def debug_mode():
    """Guess if we are in debug mode, useful to display runtime errors."""
    if '--debug' in sys.argv or '-d' in sys.argv:
        return True

    return False


def run_commands(ctx: Context, extra_args: tuple, tty: bool):
    """Run commands for a specific alias"""
    commands = ctx.obj['STAKKR'].get_config()['aliases'][ctx.command.name]['exec']
    for command in commands:
        user = command['user'] if 'user' in command else 'root'
        container = command['container']
        args = command['args'] + list(extra_args) if extra_args is not None else []

        ctx.invoke(exec_cmd, user=user, container=container, command=args, tty=tty)


def main():
    """Call the CLI Script."""
    try:
        for alias, conf in get_aliases().items():
            if conf is None:
                continue

            cmd_help = conf['description'] if 'description' in conf else 'No description'

            @stakkr.command(help=cmd_help, name=alias)
            @click.option('--tty/--no-tty', '-t/ ', is_flag=True, default=True, help="Use a TTY")
            @click.argument('extra_args', required=False, nargs=-1, type=click.UNPROCESSED)
            @click.pass_context
            def _f(ctx: Context, extra_args: tuple, tty: bool):
                """See command Help."""
                run_commands(ctx, extra_args, tty)

        stakkr(obj={})
    except Exception as error:
        msg = click.style(r""" ______ _____  _____   ____  _____
|  ____|  __ \|  __ \ / __ \|  __ \
| |__  | |__) | |__) | |  | | |__) |
|  __| |  _  /|  _  /| |  | |  _  /
| |____| | \ \| | \ \| |__| | | \ \
|______|_|  \_\_|  \_\\____/|_|  \_\

""", fg='yellow')
        msg += click.style('{}'.format(error), fg='red')
        print(msg + '\n', file=sys.stderr)

        if debug_mode() is True:
            raise error

        sys.exit(1)


if __name__ == '__main__':
    main()



# coding: utf-8
"""
Files Utils to find dir, files, etc.

Such as : static files locations or directories location
"""

from os import getcwd, listdir
from os.path import dirname, realpath


def get_lib_basedir():
    """Return the base directory of stakkr, where all files are, to read services and config."""
    return dirname(realpath(__file__))


def get_dir(directory: str):
    """Detect if stakkr is a package or a clone and gives the right path for a directory."""
    return '{}/{}'.format(get_lib_basedir(), directory)


def get_file(directory: str, filename: str):
    """Detect if stakkr is a package or a clone and gives the right path for a file."""
    return get_dir(directory) + '/' + filename.lstrip('/')


def find_project_dir():
    """Determine the project base dir, by searching a stakkr.yml file"""
    path = getcwd()
    while True:
        files = listdir(path)
        if 'stakkr.yml' in files:
            return path

        new_path = dirname(path)
        if new_path == path:
            raise FileNotFoundError('Could not find config file (stakkr.yml)')
        path = new_path
